unit Struktteile;

interface
uses Graphics,Classes, SysUtils, Windows;

const MAX_ELEMENTS = 100;

type
  TElement = class;

  TDisplayOptions = class(TObject)
    Root: TElement;
    Selected : array [1..MAX_ELEMENTS] of TElement;
    SelectedPos : TRect;
    Minimized : array [1..MAX_ELEMENTS] of TElement;
    MovePos : TElement;
    MoveBlockAfter : boolean;
    Color : boolean;
    Pfeillaenge : integer;
    Printing : boolean;
    anzSelected,anzMinimized : integer;
    function  IsSelected(e : TElement): boolean;
    function  getLastSelection(): TElement;
    function  IsMinimized(e : TElement): boolean;
    procedure ChangeMinimize(e : TElement);
    procedure Select(e : TElement);
    procedure DeselectAll();
    constructor Create();
  end;

  TElement = class(TObject)
    public
    name : String;
    next : TElement;
    movable : boolean;
    constructor laden(f : TFileStream); virtual;
    constructor copy(e : TElement);  virtual;
    procedure zeichnen(cv : TCanvas; pos : TPoint; Options: TDisplayOptions); virtual; abstract;
    procedure insert(e : TElement); virtual;
    procedure delete(e : TElement); virtual;
    procedure remove(e : TElement); virtual;
    function  findPrevious(e : TElement):TElement; virtual;
    procedure speichern(f : TFileStream); virtual; abstract;
    procedure exportieren(ss : PChar; einr : string); virtual; abstract;
    function  getSize(cv : TCanvas; Options: TDisplayOptions) : TSize; virtual; abstract;
    function  getSelection(cv : TCanvas; pos: TPoint; Options: TDisplayOptions; x,y : integer): TElement; virtual;
  end;

  TAnweisung = class(TElement)
    public
    constructor create(t : String);
    constructor laden(f : TFileStream); override;
    procedure speichern(f : TFileStream); override;
    procedure exportieren(ss : PChar; einr : string); override;
    procedure zeichnen(cv : TCanvas; pos : TPoint; Options: TDisplayOptions); override;
    function    getSize(cv : TCanvas; Options: TDisplayOptions) : TSize; override;
  end;

  TBlock = class(TElement)
    public
    elemente : TElement;
    visible  : boolean;
    showBegin : boolean;

    constructor create(t:String);
    constructor copy(e : TElement); override;
    constructor laden(f : TFileStream); override;
    procedure insert(e : TElement); override;
    procedure delete(e : TElement); override;
    procedure remove(e : TElement); override;
    function  findPrevious(e : TElement):TElement; override;
    procedure speichern(f : TFileStream); override;
    procedure exportieren(ss : PChar; einr : string); override;
    procedure zeichnen(cv : TCanvas; pos : TPoint; Options: TDisplayOptions); override;
    function  getSize(cv : TCanvas; Options: TDisplayOptions) : TSize; override;
    function  getSelection(cv : TCanvas; pos: TPoint; Options: TDisplayOptions; x,y : integer): TElement; override;
  end;

  TInvisibleBlock = class(TBlock)
    constructor create(t:String);
  end;

  TBedingung = class(TElement)
    public
    Block     : array[1..2] of TBlock;
    AnzBlocks : integer;

    constructor create(b : String);
    constructor copy(e : TElement); override;
    procedure delete(e : TElement); override;
    procedure remove(e : TElement); override;
    function getBedSize(cv : TCanvas) : TSize; virtual;
    function findPrevious(e : TElement):TElement; override;
    function getSelection(cv : TCanvas; pos: TPoint; Options: TDisplayOptions; x,y : integer): TElement; override;
    function getBlockPos(cv : TCanvas; pos: TPoint; Options: TDisplayOptions; nr : integer): TPoint; virtual; abstract;
  end;

  TIf = class(TBedingung)
    public

    constructor create(b : String);
    constructor laden(f : TFileStream); override;
    procedure speichern(f : TFileStream); override;
    procedure exportieren(ss : PChar; einr : string); override;
    procedure zeichnen(cv : TCanvas; pos : TPoint; Options: TDisplayOptions); override;
    function getSize(cv : TCanvas; Options: TDisplayOptions) : TSize; override;
    function getBlockPos(cv : TCanvas; pos: TPoint; Options: TDisplayOptions; nr : integer): TPoint; override;
  end;

  TWhile = class(TBedingung)
    public

    constructor create(b : String);
    constructor laden(f : TFileStream); override;
    procedure speichern(f : TFileStream); override;
    procedure exportieren(ss : PChar; einr : string); override;
    procedure zeichnen(cv : TCanvas; pos : TPoint; Options: TDisplayOptions); override;
    function getSize(cv : TCanvas; Options: TDisplayOptions) : TSize; override;
    function getBlockPos(cv : TCanvas; pos: TPoint; Options: TDisplayOptions; nr : integer): TPoint; override;
  end;

  TFor = class(TWhile)
    public
    constructor create(b : String);
    procedure speichern(f : TFileStream); override;
    procedure exportieren(ss : PChar; einr : string); override;
  end;


  TDoWhile = class(TWhile)
    public
    typDoW : boolean;

    constructor create(b : String);
    procedure speichern(f : TFileStream); override;
    procedure exportieren(ss: PChar; einr : string); override;
    procedure zeichnen(cv : TCanvas; pos : TPoint; Options: TDisplayOptions); override;
    function getSize(cv : TCanvas; Options: TDisplayOptions) : TSize; override;
    function getBlockPos(cv : TCanvas; pos: TPoint; Options: TDisplayOptions;  nr : integer): TPoint; override;
  end;

  TRepeat = class(TDoWhile)
    constructor create(b :String);
    constructor laden(f : TFileStream); override;
    procedure speichern(f : TFileStream); override;
    procedure exportieren(ss : PChar; einr : string); override;
  end;


implementation
type TRichtung = (RUNTER,HOCH,RECHTS,LINKS,KEINE);

// --------------------------
// Allgemeines
// --------------------------

constructor TDisplayOptions.Create();
var i : integer;
begin
  for i := 1 to MAX_ELEMENTS do
  begin
    Selected[i] := nil;
    Minimized[i] := nil;
  end;
  MovePos := nil;
  MoveBlockAfter := false;
  anzSelected := 0;
  anzMinimized := 0;
  color := true;
  Pfeillaenge := 6;
  Printing := false;
  SelectedPos.left := 0;
  SelectedPos.right := 0;
  SelectedPos.top := 0;
  SelectedPos.bottom := 0;

end;

function  TDisplayOptions.IsSelected;
var b : boolean;
    i : integer;
begin
  b := false;
  for i := 1 to anzSelected do
  begin
    if e = Selected[i] then b := true;
  end;
  result := b;
end;

function TDisplayOptions.getLastSelection;
begin
  if anzSelected=0 then
    result := nil
  else
    result := Selected[anzSelected];
end;

function  TDisplayOptions.IsMinimized;
var b : boolean;
    i : integer;
begin
  b := false;
  for i := 1 to anzMinimized do
  begin
    if e = Minimized[i] then b := true;
  end;
  result := b;
end;

procedure TDisplayOptions.ChangeMinimize(e : TElement);
var b : boolean;
    i : integer;
begin
  b := false;
  for i := 1 to anzMinimized do
  begin
    if b then Minimized[i-1] := Minimized[i];
    if e = Minimized[i] then b := true;
  end;
  if b then
  begin
    Minimized[anzMinimized] := nil;
    anzMinimized := anzMinimized -1;
  end
  else
  begin
    anzMinimized := anzMinimized +1;
    Minimized[anzMinimized]:=e;
  end;
end;

procedure TDisplayOptions.Select(e : TElement);
var b : boolean;
    i : integer;
begin
  b := false;
  for i := 1 to anzSelected do
  begin
    if e = Selected[i] then b := true;
  end;
  if not b then
  begin
    anzSelected := anzSelected +1;
    Selected[anzSelected]:=e;
  end;
end;

procedure TDisplayOptions.DeselectAll();
var i : integer;
begin
  for i := 1 to anzSelected do
  begin
    Selected[i]:= nil;
  end;
  anzSelected := 0;
end;

procedure SetSelectionColor(cv : TCanvas);
begin
  cv.pen.color := clRed;
  cv.Font.color := clRed;
end;

procedure SetNormalColor(cv : TCanvas);
begin
  cv.pen.color := clBlack;
  cv.Font.color := clBlack;
end;

procedure ZeichnePfeil(cv : TCanvas; r : TRichtung; x1,y1,x2,y2 : integer; highLight:boolean; Printing:boolean);
var grSpitze:integer;
    col : TColor;
begin
  col := cv.brush.color;
  cv.brush.color:=clBlack;
  if (Printing) then
    grSpitze := 10
  else
  begin
    if (highlight) then
    begin
      cv.pen.color := clRed;
      cv.brush.color := clRed;
    end;
    grSpitze := 2;
  end;
  cv.moveTo(x1,y1);
  cv.lineTo(x2,y2);
  case r of
  RUNTER: cv.Polygon([Point(x2-grSpitze,y2-grSpitze),Point(x2+grSpitze,y2-grSpitze),Point(x2,y2)]);
  RECHTS: cv.Polygon([Point(x2-grSpitze,y2-grSpitze),Point(x2-grSpitze,y2+grSpitze),Point(x2,y2)]);
  KEINE : cv.lineTo(x2,y2+1);
  end;
  if (highlight) then cv.pen.color := clBlack;
  cv.brush.color := col;
end;

function getString(f : TFileStream):string;
var s: array[0..255] of char;
    s1: array[0..2] of char;
begin
  strcopy(s,'');
  strcopy(s1,' ');
  repeat
    f.readBuffer(s1,1);
    if (s1 <> #1) then strcat(s,s1);
  until s1 = #1;
  result := s;
end;


// --------------------------
//         TElement
// --------------------------

constructor TElement.Laden(f : TFileStream);
var s: array [0..2] of char;
begin
  if (f.Read(s,2)<>2) then   strcopy(s,#2);
  case s[0] of
    #2 :  next := nil;
    'A' : next := TAnweisung.Laden(f);
    'B' : next := TBlock.Laden(f);
    'D' : next := TDoWhile.Laden(f);
    'F' : next := TFor.Laden(f);
    'I' : next := TIF.Laden(f);
    'R' : next := TRepeat.Laden(f);
    'W' : next := TWhile.Laden(f);
  end;
  movable:=true;
end;

constructor TElement.Copy(e : TElement);
begin
  if (e.next is TAnweisung) then next := TAnweisung.copy(e.next);
  if (e.next is TBlock) then next := TBlock.copy(e.next);
  if (e.next is TIf) then next := TIf.copy(e.next);
  if (e.next is TRepeat) then next := TRepeat.copy(e.next)
  else begin
  if (e.next is TDoWhile) then next := TDoWhile.copy(e.next)
  else begin
  if (e.next is TFor) then next := TFor.copy(e.next)
  else begin
  if (e.next is TWhile) then next := TWhile.copy(e.next);
  end; end; end;
  name := e.name;
  movable:= e.movable;
end;

function TElement.GetSelection;
var g : TSize;
begin
  g := getSize(cv,Options);
  if (x >= pos.x - g.cx div 2) and (x<= pos.x +g.cx div 2) and
     (y >= pos.y) and (y<= pos.y +g.cy) then
  begin
    result := self;
    Options.MoveBlockAfter := false;
  end
  else
  begin
    pos.y := pos.y + g.cy;
    if next <> nil then result := next.getSelection(cv,pos,Options,x,y)
                   else
                   begin
                     result := nil;
                     Options.MoveBlockAfter := false;
                   end;
  end;
end;

procedure TElement.Insert(e : TElement);
begin
  e.next := next;
  next := e;
end;

procedure TElement.Delete(e : TElement);
begin
  if next <> nil then next.delete(e);
  if e = next then
  begin
    next := e.next;
  end;
end;

procedure TElement.Remove(e : TElement);
begin
  if e = next then
  begin
    next := e.next;
  end
  else
    if next <> nil then next.remove(e);
end;


function  TElement.findPrevious(e : TElement):TElement;
begin
  if next = e then result := self
              else if (next = nil) then result := nil
                                   else result := next.findPrevious(e);
end;



// --------------------------
//         TAnweisung
// --------------------------

constructor TAnweisung.Create(t: String);
begin
  name := t;
  movable := true;
  next := nil;
end;

constructor TAnweisung.Laden(f : TFileStream);
begin
  name := getString(f);
  inherited;
end;


procedure TAnweisung.Speichern(f:TFileStream);
var s:array[0..100] of char;
begin
  strcopy(s,'A:');
  strcat(s,PChar(name));
  strcat(s,#1);
  f.WriteBuffer(s,strlen(s));
  if (next<> nil) then next.Speichern(f);
end;

procedure TAnweisung.Exportieren;
var s:array[0..100] of char;
begin
  strcopy(s,PChar(einr));
  strcat(s,PChar(name));
  if not (s[strlen(s)]=';') then strcat(s,';');
  strcat(s,#13+#10);
  strcat(ss,s);
  if (next<> nil) then next.exportieren(ss,einr);
end;

procedure TAnweisung.zeichnen;
var groesse : TSize;
    oldColor : TColor;
    r : TRect;
begin
  groesse := getSize(cv,Options);
  if Options.Printing then groesse.cx := groesse.cx - 2*Options.Pfeillaenge;
  oldColor := cv.brush.Color;
  cv.brush.Color := clWhite;
  r.left :=pos.x-groesse.cx div 2; r.top := pos.y;
  r.right:=pos.x+groesse.cx div 2; r.bottom := pos.y + groesse.cy - Options.Pfeillaenge;
  if Options.IsSelected(self) and options.color then
  begin
    setSelectionColor(cv);
    Options.SelectedPos:=r;
  end;
  cv.fillrect(r);
  cv.rectangle(r.left,r.top,r.Right,r.bottom);
  cv.TextOut(pos.x-groesse.cx div 2 + 5, pos.y+5, name);
  ZeichnePfeil(cv,RUNTER,pos.x,pos.y+groesse.cy-Options.Pfeillaenge-1,pos.x,pos.y+groesse.cy-1,Options.MovePos=self,Options.Printing);
  pos.y := pos.y + groesse.cy;
  setNormalColor(cv);
  cv.brush.Color:=oldColor;
  if next <> nil then next.zeichnen(cv,pos,Options);
end;

function TAnweisung.getSize;
var groesse : TSize;
begin
  groesse := cv.TextExtent(name);
  groesse.cx := groesse.cx + 12;
  groesse.cy := groesse.cy + 11 + Options.Pfeillaenge;
  if Options.Printing then groesse.cx := groesse.cx + 2*Options.Pfeillaenge;
  result := groesse;
end;

// --------------------------
//         TBlock
// --------------------------

constructor TBlock.Create(t: String);
begin
  name := t;
  movable:=true;
  next := nil;
  elemente := nil;
  visible := true;
  ShowBegin := true;
end;

constructor TBlock.copy(e : TElement);
// var b : TElement;
begin
  inherited;
//  b:= (e as TBlock).elemente;
//  if b<>nil then b:= nil;
  visible := (e as TBlock).visible;
  ShowBegin := (e as TBlock).ShowBegin;
  if ((e as TBlock).elemente is TAnweisung) then elemente := TAnweisung.copy((e as TBlock).elemente);
  if ((e as TBlock).elemente is TBlock) then elemente := TBlock.copy((e as TBlock).elemente);
  if ((e as TBlock).elemente is TIf) then elemente := TIf.copy((e as TBlock).elemente);
  if ((e as TBlock).elemente is TRepeat) then elemente := TRepeat.copy((e as TBlock).elemente)
  else begin
  if ((e as TBlock).elemente is TDoWhile) then elemente := TDoWhile.copy((e as TBlock).elemente)
  else begin
  if ((e as TBlock).elemente is TFor) then elemente := TFor.copy((e as TBlock).elemente)
  else begin
  if ((e as TBlock).elemente is TWhile) then elemente := TWhile.copy((e as TBlock).elemente);
  end; end; end;
end;

constructor TBlock.Laden(f : TFileStream);
var s: array [0..2] of char;
begin
  name := getString(f);
  visible := true;
  ShowBegin := true;
  if (f.Read(s,2)<>2) then strcopy(s,#2);
  case s[0] of
    #2 :  elemente := nil;
    'A' : elemente := TAnweisung.Laden(f);
    'B' : elemente := TBlock.Laden(f);
    'D' : elemente := TDoWhile.Laden(f);
    'F' : elemente := TFor.Laden(f);
    'I' : elemente := TIF.Laden(f);
    'R' : elemente := TRepeat.Laden(f);
    'W' : elemente := TWhile.Laden(f);
  end;
  inherited;
end;

constructor TInvisibleBlock.Create(t: String);
begin
  inherited;
  visible := false;
end;

procedure TBlock.Insert(e : TElement);
begin
  e.next := elemente;
  elemente := e;
end;

procedure TBlock.Delete;
var e2: TElement;
begin
  inherited;
  if (e = self) then
  begin
    // alle untergeordneten freigeben
    while elemente <> nil do
    begin
      e2 := elemente.next;
      elemente.delete(elemente);
      elemente.free;
      elemente := e2;
    end;
  end
  else
    if elemente <> nil then
    begin
      if elemente = e then elemente := e.next
                      else elemente.delete(e);
    end;
end;

procedure TBlock.Remove;
begin
  inherited;
  if (elemente<>nil) then
  begin
    if (elemente=e) then elemente := e.next
                    else elemente.remove(e);
  end;
end;


function  TBlock.findPrevious(e : TElement):TElement;
var erg : TElement;
begin
  erg := nil;
  if next = e then erg := self
  else if elemente = e then erg := self
  else
  begin
    if (elemente <> nil) then erg := elemente.findPrevious(e);
    if (erg = nil) and (next <> nil) then erg:= next.findPrevious(e);
  end;
  result := erg;
end;

procedure TBlock.Speichern(f:TFileStream);
var s:array[0..100] of char;
begin
  strcopy(s,'B:');
  strcat(s,PChar(name));
  strcat(s,#1);
  f.WriteBuffer(s,strlen(s));
  if (elemente<>nil) then elemente.Speichern(f);
  strcopy(s,#2);
  strcat(s,#2);
  f.WriteBuffer(s,2);
  if (next<> nil) then next.Speichern(f);
end;

procedure TBlock.Exportieren;
var s:array[0..100] of char;
    einr2: string;
begin
  strcopy(s,PChar(einr));
  einr2 := einr;
  if movable then
  begin
    strcat(s,#13+#10);
    strcat(s,PChar(einr));
    strcat(s,'// ') ;
    strcat(s,PChar(name));
    strcat(s,#13+#10);
    strcat(ss,s);
  end
  else
  begin
    einr2 := einr2+'  ';
    if (ShowBegin) then
    begin
      strcat(s,'begin');
      strcat(s,#13+#10);
      strcat(ss,s);
    end;
  end;
  if elemente <> nil then elemente.exportieren(ss,einr2);
  if (movable) then
  begin
    strcopy(s,PChar(einr));
    strcat(s,'// Ende ');
    strcat(s,PChar(name));
    strcat(s,#13+#10); strcat(s,#13+#10);
    strcat(ss,s);
  end
  else
  begin
    if (ShowBegin) then
    begin
      strcopy(s,PChar(einr));
      strcat(s,'end');
      strcat(ss,s);
    end
  end;
  if (next<> nil) then next.Exportieren(ss,einr);
end;

function TBlock.GetSelection;
var e,e2:TElement;
begin
  e := inherited getSelection(cv,pos,Options,x,y);
  if (e = self) and (elemente <> nil) and not Options.IsMinimized(self) then
  begin
    if visible then
    begin
      pos.y := pos.y + Options.Pfeillaenge+1;
    end;
    e2 := elemente.getSelection(cv,pos,Options,x,y);
    if e2 <> nil then e := e2;
    if (e = self) and (Movable) and (pos.y < y) then
      Options.MoveBlockAfter:= true;
  end;
  result := e;
end;

procedure TBlock.zeichnen;
var groesse : TSize;
    pos1 : TPoint;
begin
  groesse := getSize(cv,Options);
  if Options.Printing then groesse.cx := groesse.cx - Options.Pfeillaenge;
  if visible then
  begin
    if options.color then cv.brush.Color := cv.brush.Color - $00000606;
    cv.pen.style := psDot;
    if Options.IsSelected(self) and options.color then
    begin
      setSelectionColor(cv);
      Options.SelectedPos.topleft:=Point(pos.x-groesse.cx, pos.y);
      Options.SelectedPos.bottomright:=Point(pos.x+groesse.cx, pos.y+groesse.cy);
    end;
    if (cv.pen.color <> clRed) and options.color then cv.pen.color := clGray;
    cv.pen.width := 1;
    cv.polygon([Point(pos.x-groesse.cx div 2, pos.y),
                Point(pos.x+groesse.cx div 2, pos.y),
                Point(pos.x+groesse.cx div 2, pos.y + groesse.cy-Options.Pfeillaenge-1),
                Point(pos.x-groesse.cx div 2, pos.y + groesse.cy-Options.Pfeillaenge-1)
               ]);
    if Options.Printing then cv.pen.width := 3;
    if cv.pen.color <> clRed then cv.pen.color := clBlack;
    cv.pen.style := psSolid;
    if Options.IsMinimized(Self) then
      cv.TextOut(pos.x-groesse.cx div 2 + 5, pos.y+5, name);
    if movable then
      ZeichnePfeil(cv,RUNTER,pos.x,pos.y+groesse.cy-Options.Pfeillaenge-1,pos.x,pos.y+groesse.cy-1,(Options.MovePos=self) and (Options.MoveBlockAfter),Options.Printing)
    else
      ZeichnePfeil(cv,KEINE,pos.x,pos.y+groesse.cy-Options.Pfeillaenge-1,pos.x,pos.y+groesse.cy-1,(Options.MovePos=self) and (Options.MoveBlockAfter),Options.Printing)
  end;

  pos1 := pos;
  if not Options.IsMinimized(self) then
  if (elemente <> nil)  then
  begin
    if visible then
    begin
      ZeichnePfeil(cv,RUNTER,pos.x,pos.y,pos.x,pos.y+Options.Pfeillaenge,(Options.MovePos=self) and not (Options.MoveBlockAfter),Options.Printing);
      pos1.y := pos1.y + Options.Pfeillaenge+1;
    end;
    setNormalColor(cv);
    elemente.zeichnen(cv,pos1,Options)
  end
  else
  begin
    ZeichnePfeil(cv,RUNTER,pos.x,pos.y,pos.x,pos.y+groesse.cy-Options.Pfeillaenge-1,(Options.MovePos=self) and not (Options.MoveBlockAfter),Options.Printing);
  end;
  if visible and options.color then
  begin
    cv.brush.Color := cv.brush.Color + $00000606;
  end;
  setNormalColor(cv);
  pos.y := pos.y + groesse.cy;
  if (next <> nil) and (self <> options.root) then next.zeichnen(cv,pos,Options);
end;

function TBlock.getSize;
var groesse,gr1 : TSize;
    actElement : TElement;
begin
  if Options.IsMinimized(self) then
  begin
    groesse := cv.TextExtent(name);
    groesse.cx := groesse.cx + 10;
    groesse.cy := groesse.cy + 11 + Options.Pfeillaenge;
  end
  else
  begin
    groesse.cx := 0;
    groesse.cy := 0;
    actElement := elemente;
    while (actElement <> nil) do
    begin
      gr1 := actElement.getSize(cv,Options);
      if (groesse.cx < gr1.cx) then groesse.cx := gr1.cx;
      groesse.cy := groesse.cy + gr1.cy;
      actElement := actElement.next;
    end;
    if groesse.cx < 40 then groesse.cx := 40;
    if (Options.Printing) and (groesse.cx <100) then groesse.cx := 100;
    if groesse.cy < 20 then groesse.cy := 20;

    groesse.cx := groesse.cx + 6;
    if visible then groesse.cy := groesse.cy + 2*Options.Pfeillaenge+3;
  end;
  if Options.Printing then groesse.cx := groesse.cx + Options.Pfeillaenge;
  result := groesse;
end;

// --------------------------
//         TBedingung
// --------------------------

constructor TBedingung.Create(b: String);
begin
  name := b;
  anzBlocks := 0;
  movable   := true;
  next      := nil;
  Block[1]  := nil;
  Block[2]  := nil;
end;

constructor TBedingung.copy(e : TElement);
var b : TBedingung;
begin
  inherited;
  b := (e as TBedingung);
  anzBlocks := b.anzBlocks;
  if b.Block[1] <> nil then  Block[1] := TBlock.copy(b.Block[1])
                       else  Block[1] := nil;
  if b.Block[2] <> nil then  Block[2] := TBlock.copy(b.Block[2])
                       else  Block[2] := nil;
end;

procedure TBedingung.Delete;
var i: integer;
begin
  inherited;
  if (e = self) then
  begin
    // alle untergeordneten freigeben
    for i := 1 to anzBlocks do
    begin
      block[i].delete(block[i]);
      block[i].free;
    end;
  end
  else
  begin
    for i := 1 to anzBlocks do
    begin
      block[i].delete(e);
    end;
  end;
end;

procedure TBedingung.Remove;
var i : integer;
begin
  inherited;
  for i := 1 to anzBlocks do
  begin
    block[i].remove(e);
  end;
end;


function TBedingung.getBedSize(cv : TCanvas):TSize;
var groesse : TSize;
begin
  groesse := cv.TextExtent(name);
  groesse.cy := groesse.cy + groesse.cx div 3+2;
  groesse.cx := 3*groesse.cy+2;
  result := groesse;
end;


function TBedingung.GetSelection;
var e,e2:TElement;
    i : integer;
begin
  e := inherited getSelection(cv,pos,Options,x,y);
  if (e = self) then
  begin
    for i := 1 to anzBlocks do
    begin
      e2 := block[i].getSelection(cv,getBlockPos(cv,pos,Options,i),Options,x,y);
      if e2 <> nil then e := e2;
    end;
  end;
  result := e;
end;

function  TBedingung.findPrevious(e : TElement):TElement;
var erg : TElement;
begin
  erg := nil;
  if next = e then erg := self
  else if Block[1] = e then erg := self
  else if Block[2] = e then erg := self
  else
  begin
    if (Block[1] <> nil) then erg := Block[1].findPrevious(e);
    if (Block[2] <> nil) and (erg = nil) then erg := Block[2].findPrevious(e);
    if (erg = nil) and (next <> nil) then erg:= next.findPrevious(e);
  end;
  result := erg;
end;

// --------------------------
//         TIf
// --------------------------

constructor TIf.Create(b: String);
begin
  inherited;
  Block[1] := TBlock.Create('Ja-Block');
  Block[2] := TBlock.Create('Nein-Block');
  anzBlocks := 2;
  Block[1].movable:=false;
  Block[2].movable:=false;
end;

constructor TIf.Laden(f: TFileStream);
var s : array[0..2] of char;
begin
  anzBlocks := 2;
  name := getString(f);
  f.Read(s,2);
  Block[1] := TBlock.Laden(f);
  f.Read(s,2);
  Block[2] := TBlock.Laden(f);
  Block[1].movable:=false;
  Block[2].movable:=false;
  inherited;
end;

procedure TIf.Speichern(f:TFileStream);
var s:array[0..100] of char;
begin
  strcopy(s,'I:');
  strcat(s,PChar(name));
  strcat(s,#1);
  f.WriteBuffer(s,strlen(s));
  block[1].Speichern(f);
  strcopy(s,#2);
  strcat(s,#2);
  f.WriteBuffer(s,2);
  block[2].Speichern(f);
  f.WriteBuffer(s,2);
  if (next<> nil) then next.Speichern(f);
end;

procedure TIf.Exportieren;
var s:array[0..100] of char;
begin
  strcopy(s,PChar(einr));
  strcat(s,'if ');
  strcat(s,PChar(name));
  strcat(s,' then');
  strcat(s,#13+#10);
  strcat(ss,s);
  block[1].exportieren(ss,einr);
  if (block[2].elemente<>NIL) then
  begin
    strcopy(s,#13+#10);
    strcat(s,PChar(einr));
    strcat(s,'else');
    strcat(s,#13+#10);
    strcat(ss,s);
    block[2].exportieren(ss,einr);
  end;
  strcopy(s,';');
  strcat(s,#13+#10);
  strcat(ss,s);
  if (next<> nil) then next.Exportieren(ss,einr);
end;

procedure TIf.zeichnen;
var groesse,gr1 : TSize;
    post,pose   : TPoint;
    oldColor    : TColor;
begin
  groesse := getBedSize(cv);
  gr1     := cv.TextExtent(name);
  post    := getBlockPos(cv,pos,Options,1);
  pose    := getBlockPos(cv,pos,Options,2);

  if Options.IsSelected(self) and options.color then
  begin
    setSelectionColor(cv);
    Options.SelectedPos.topleft:=Point(pos.x-groesse.cx, pos.y);
    Options.SelectedPos.bottomright:=Point(pos.x+groesse.cx, pos.y+groesse.cy);
  end;

  // Ja-Nein Pfeile zeichnen
  cv.moveTo(post.x,pos.y+groesse.cy-5);
  cv.lineTo(pos.x,pos.y+groesse.cy div 2);
  cv.lineTo(pose.x,pos.y+groesse.cy-5);
  ZeichnePfeil(cv,RUNTER,post.x,pos.y+groesse.cy-5,post.x,pos.y+groesse.cy+3+Options.Pfeillaenge,false,Options.Printing);
  ZeichnePfeil(cv,RUNTER,pose.x,pos.y+groesse.cy-5,pose.x,pos.y+groesse.cy+3+Options.Pfeillaenge,false,Options.Printing);
  if Options.Printing then
  begin
    cv.TextOut(post.x - cv.TextExtent('ja').cx - 10,pos.y + groesse.cy +3+Options.Pfeillaenge - cv.TextExtent('ja').cy, 'ja');
    cv.TextOut(pose.x + 10,pos.y + groesse.cy +3+Options.Pfeillaenge -  cv.TextExtent('ja').cy, 'nein');
  end
  else
  begin
    cv.TextOut(post.x - cv.TextExtent('ja').cx - 2,pos.y + groesse.cy +3+Options.Pfeillaenge - cv.TextExtent('ja').cy, 'ja');
    cv.TextOut(pose.x + 2,pos.y + groesse.cy +3+Options.Pfeillaenge -  cv.TextExtent('ja').cy, 'nein');
  end;

  // Bedingung zeichnen
  oldColor := cv.brush.Color;
  cv.brush.Color := clWhite;
  cv.polygon([Point(pos.x,pos.y),
              Point(pos.x+groesse.cx div 2, pos.y+groesse.cy div 2),
              Point(pos.x, pos.y+groesse.cy),
              Point(pos.x-groesse.cx div 2, pos.y+groesse.cy div 2)]);
  cv.TextOut(pos.x - gr1.cx div 2, pos.y + groesse.cy div 2 - gr1.cy div 2,name);
  cv.brush.Color := oldColor;

  // Blocks zeichnen
  Block[1].zeichnen(cv,post,Options);
  if Options.IsSelected(self) and options.color then setSelectionColor(cv);
  Block[2].zeichnen(cv,pose,Options);
  if Options.IsSelected(self) and options.color then setSelectionColor(cv);

  // Pfeilabschluss
  if (Options.MovePos=self) then cv.Pen.Color := clRed;
  post.y := post.y + Block[1].getSize(cv,Options).cy;
  pose.y := pose.y + Block[2].getSize(cv,Options).cy;
  if (post.y > pose.y) then
  begin
    cv.moveTo(pose.x,pose.y); cv.lineTo(pose.x,post.y);
    cv.lineTo(post.x,post.y);
    pos.y := post.y;
  end
  else
  begin
    cv.moveTo(post.x,post.y); cv.lineTo(post.x,pose.y);
    cv.lineTo(pose.x,pose.y);
    pos.y := pose.y;
  end;
  ZeichnePfeil(cv,RUNTER,pos.x,pos.y+1,pos.x,pos.y+1+Options.Pfeillaenge,Options.MovePos=self,Options.Printing);
  setNormalColor(cv);


  pos.y  := pos.y + Options.Pfeillaenge +1;

  if next <> nil then next.zeichnen(cv,pos,Options);
end;



function TIf.getSize;
var groesse,grt,gre : TSize;
begin
  groesse := getBedSize(cv);
  grt := Block[1].getSize(cv,Options);
  gre := Block[2].getSize(cv,Options);
  if (grt.cy > gre.cy) then
    groesse.cy := groesse.cy + 10 + grt.cy
  else
    groesse.cy := groesse.cy + 10 + gre.cy ;

  groesse.cy := groesse.cy + 2*Options.Pfeillaenge -5;
  groesse.cx := groesse.cx + 6;
  if (grt.cx+gre.cx+10 > groesse.cx) then groesse.cx := grt.cx+gre.cx+10;
  result := groesse;
end;

function TIf.GetBlockPos;
var breite : integer;
begin
  pos.y  := pos.y + getBedSize(cv).cy + 4 + Options.Pfeillaenge;
  breite := (Block[2].getSize(cv,Options).cx + Block[1].getSize(cv,Options).cx + 6) div 2;
  if (nr = 1 ) then pos.x  := pos.x - breite + Block[1].getSize(cv,Options).cx div 2
               else pos.x  := pos.x + breite - Block[2].getSize(cv,Options).cx div 2;
  result := pos;
end;

// --------------------------
//         TWhile
// --------------------------

constructor TWhile.Create(b: String);
begin
  inherited;
  Block[1] := TBlock.Create('While-Block');
  Block[1].movable:=false;
  anzBlocks := 1;
end;

constructor TWhile.Laden(f: TFileStream);
var s : array[0..2] of char;
begin
  anzBlocks := 1;
  name := getString(f);
  f.Read(s,2);
  Block[1] := TBlock.Laden(f);
  Block[1].movable:=false;
  inherited;
end;

procedure TWhile.Speichern(f:TFileStream);
var s:array[0..100] of char;
begin
  strcopy(s,'W:');
  strcat(s,PChar(name));
  strcat(s,#1);
  f.WriteBuffer(s,strlen(s));
  block[1].Speichern(f);
  strcopy(s,#2);
  strcat(s,#2);
  f.WriteBuffer(s,2);
  if (next<> nil) then next.Speichern(f);
end;

procedure TWhile.Exportieren;
var s:array[0..100] of char;
begin
  strcopy(s,PChar(einr));
  strcat(s,'while ');
  strcat(s,PChar(name));
  strcat(s,' do');
  strcat(s,#13+#10);
  strcat(ss,s);
  block[1].exportieren(ss,einr);
  strcopy(s,';');
  strcat(s,#13+#10);
  strcat(ss,s);
  if (next<> nil) then next.Exportieren(ss,einr);
end;


procedure TWhile.zeichnen;
var groesse,gr1,gr2 : TSize;
    oldColor    : TColor;
begin
  groesse := getBedSize(cv);
  gr1     := cv.TextExtent(name);

  // Bedingung zeichnen
  oldColor := cv.brush.Color;
  if Options.IsSelected(self) and options.color then
  begin
    setSelectionColor(cv);
    Options.SelectedPos.topleft:=Point(pos.x-groesse.cx, pos.y);
    Options.SelectedPos.bottomright:=Point(pos.x+groesse.cx, pos.y+groesse.cy);
  end;

  cv.brush.Color := clWhite;
  cv.polygon([Point(pos.x,pos.y),
              Point(pos.x+groesse.cx div 2, pos.y+groesse.cy div 2),
              Point(pos.x, pos.y+groesse.cy),
              Point(pos.x-groesse.cx div 2, pos.y+groesse.cy div 2)]);
  cv.TextOut(pos.x - gr1.cx div 2, pos.y + groesse.cy div 2 - gr1.cy div 2,name);
  cv.brush.Color := oldColor;

  // Block zeichnen
  ZeichnePfeil  (cv,RUNTER,pos.x,pos.y+groesse.cy,pos.x,pos.y+groesse.cy+Options.Pfeillaenge+cv.TextExtent('ja').cy,false,Options.Printing);
  Block[1].zeichnen(cv,Point(pos.x,pos.y+groesse.cy+Options.Pfeillaenge+1+cv.TextExtent('ja').cy),Options);
  if Options.IsSelected(self) and options.color then setSelectionColor(cv);

  // Pfeil rechts rum
  if (Options.MovePos=self) then cv.Pen.Color := clRed;
  gr2 := Block[1].getSize(cv,Options);
  if (groesse.cx > gr2.cx ) then gr2.cx := groesse.cx;
  cv.moveTo(pos.x+groesse.cx div 2,pos.y+groesse.cy div 2);
  cv.lineTo(pos.x+6+gr2.cx div 2,pos.y+groesse.cy div 2);
  cv.lineTo(pos.x+6+gr2.cx div 2,pos.y+groesse.cy+gr2.cy+Options.Pfeillaenge+13+cv.TextExtent('ja').cy);
  cv.lineTo(pos.x,pos.y+groesse.cy+gr2.cy+Options.Pfeillaenge+13+cv.TextExtent('ja').cy);
  ZeichnePfeil(cv,RUNTER,pos.x,pos.y+groesse.cy+gr2.cy+Options.Pfeillaenge+13+cv.TextExtent('ja').cy,
                         pos.x,pos.y+groesse.cy+gr2.cy+2*Options.Pfeillaenge+13+cv.TextExtent('ja').cy,Options.MovePos=self,Options.Printing);

  // Pfeil links hoch
  cv.moveTo(pos.x, pos.y+groesse.cy+gr2.cy+Options.Pfeillaenge+cv.TextExtent('ja').cy);
  cv.lineTo(pos.x, pos.y+groesse.cy+gr2.cy+Options.Pfeillaenge+3+cv.TextExtent('ja').cy);
  cv.lineTo(pos.x-6-gr2.cx div 2,pos.y+groesse.cy+gr2.cy+Options.Pfeillaenge+3+cv.TextExtent('ja').cy);
  cv.lineTo(pos.x-6-gr2.cx div 2,pos.y+groesse.cy div 2);
  ZeichnePfeil(cv,RECHTS,pos.x-6-gr2.cx div 2,pos.y+groesse.cy div 2,
                         pos.x-groesse.cx div 2, pos.y+groesse.cy div 2,false,Options.Printing);

  // Beschriftung der Pfeile
  cv.TextOut(pos.x - cv.TextExtent('ja').cx - 2,pos.y + groesse.cy, 'ja');
  cv.TextOut(pos.x + groesse.cx div 2 - 4,pos.y + groesse.cy div 2 - cv.TextExtent('nein').cy, 'nein');
  setNormalColor(cv);

  pos.y := pos.y + getSize(cv,Options).cy;
  if next <> nil then next.zeichnen(cv,pos,Options);
end;



function TWhile.getSize;
var groesse,grt : TSize;
begin
  groesse := getBedSize(cv);
  grt := Block[1].getSize(cv,Options);
  groesse.cy := groesse.cy + Options.Pfeillaenge*2 + grt.cy+ 14+ cv.TextExtent('ja').cy;
  groesse.cx := groesse.cx + 2*cv.TextExtent('nein').cx - 4;
  if (grt.cx > groesse.cx) then groesse.cx := grt.cx+18;
  result := groesse;
end;

function TWhile.GetBlockPos;
begin
  result := Point(pos.x,pos.y+getBedSize(cv).cy+Options.Pfeillaenge+1+cv.TextExtent('ja').cy);
end;

// --------------------------
//         TDoWhile
// --------------------------

constructor TDoWhile.Create(b: String);
begin
  inherited;
  Block[1].Name := 'DoWhile-Block';
  typDoW := true;
end;

constructor TRepeat.Create(b: String);
begin
  inherited;
  Block[1].Name := 'RepeatUntil-Block';
  typDoW := false;
  Block[1].ShowBegin := false;
end;

constructor TRepeat.Laden;
begin
  inherited;
  Block[1].ShowBegin := false;
end;


procedure TDoWhile.Speichern(f:TFileStream);
var s:array[0..100] of char;
begin
  strcopy(s,'D:');
  strcat(s,PChar(name));
  strcat(s,#1);
  f.WriteBuffer(s,strlen(s));
  block[1].Speichern(f);
  strcopy(s,#2);
  strcat(s,#2);
  f.WriteBuffer(s,2);
  if (next<> nil) then next.Speichern(f);
end;

procedure TRepeat.Speichern(f:TFileStream);
var s:array[0..100] of char;
begin
  strcopy(s,'R:');
  strcat(s,PChar(name));
  strcat(s,#1);
  f.WriteBuffer(s,strlen(s));
  block[1].Speichern(f);
  strcopy(s,#2);
  strcat(s,#2);
  f.WriteBuffer(s,2);
  if (next<> nil) then next.Speichern(f);
end;

procedure TDoWhile.Exportieren;
begin
  if (next<> nil) then next.Exportieren(ss,einr);
end;

procedure TRepeat.Exportieren;
var s:array[0..100] of char;
begin
  strcopy(s,PChar(einr));
  strcat(s,'repeat');
  strcat(s,#13+#10);
  strcat(ss,s);
  block[1].exportieren(ss,einr);
  strcopy(s,PChar(einr));
  strcat(s,'until ');
  strcat(s,PChar(name));
  strcat(s,';');
  strcat(s,#13+#10);
  strcat(ss,s);
  if (next<> nil) then next.Exportieren(ss,einr);
end;

procedure TDoWhile.zeichnen;
var groesse,gr1,gr2 : TSize;
    oldColor    : TColor;
begin
  groesse := getBedSize(cv);
  gr1     := cv.TextExtent(name);
  gr2     := Block[1].getSize(cv,Options);

  // Block zeichnen
  if Options.IsSelected(self) and options.color then
  begin
    setSelectionColor(cv);
    Options.SelectedPos.topleft:=Point(pos.x-groesse.cx, pos.y);
    Options.SelectedPos.bottomright:=Point(pos.x+groesse.cx, pos.y+groesse.cy);
  end;
  ZeichnePfeil(cv,RUNTER,pos.x,pos.y,
                         pos.x,pos.y+12,false,Options.Printing);
  Block[1].zeichnen(cv,Point(pos.x,pos.y+12),Options);
  if Options.IsSelected(self) and options.color then setSelectionColor(cv);

  // Beschriftung der Pfeile
  if typDoW then
  begin
    cv.TextOut(pos.x - cv.TextExtent('nein').cx - 2,pos.y + groesse.cy+gr2.cy+11, 'nein');
    cv.TextOut(pos.x - groesse.cx div 2 - cv.TextExtent('ja').cx+5,pos.y +gr2.cy +12 + groesse.cy div 2+1, 'ja');
  end
  else
  begin
    cv.TextOut(pos.x - cv.TextExtent('ja').cx - 2,pos.y + groesse.cy+gr2.cy+11, 'ja');
    cv.TextOut(pos.x - groesse.cx div 2 - cv.TextExtent('nein').cx+5,pos.y +gr2.cy +12 + groesse.cy div 2+1, 'nein');
  end;

  // Bedingung zeichnen
  oldColor := cv.brush.Color;
  cv.brush.Color := clWhite;
  cv.polygon([Point(pos.x,pos.y+gr2.cy+12),
              Point(pos.x+groesse.cx div 2, pos.y+gr2.cy+12+groesse.cy div 2),
              Point(pos.x, pos.y+gr2.cy+12+groesse.cy),
              Point(pos.x-groesse.cx div 2, pos.y+gr2.cy+12+groesse.cy div 2)]);
  cv.TextOut(pos.x - gr1.cx div 2, pos.y +gr2.cy +12 + groesse.cy div 2 - gr1.cy div 2,name);
  cv.brush.Color := oldColor;


  // Pfeil runter
  ZeichnePfeil(cv,RUNTER,pos.x,pos.y+groesse.cy+gr2.cy+13,
                         pos.x,pos.y+groesse.cy+gr2.cy+cv.TextExtent('ja').cy+13,Options.MovePos=self,Options.Printing);


  // Pfeil links hoch
  if (gr2.cx < groesse.cx) then gr2.cx := groesse.cx;
  cv.moveTo(pos.x-groesse.cx div 2, pos.y+groesse.cy div 2 +gr2.cy+12);
  cv.lineTo(pos.x-6-gr2.cx div 2,   pos.y+groesse.cy div 2 +gr2.cy+12);
  cv.lineTo(pos.x-6-gr2.cx div 2,pos.y+4);
  ZeichnePfeil(cv,RECHTS,pos.x-6-gr2.cx div 2,pos.y+4,
                         pos.x-1, pos.y+4,false,Options.Printing);

  setNormalColor(cv);

  pos.y := pos.y + getSize(cv,Options).cy;
  if next <> nil then next.zeichnen(cv,pos,Options);
end;



function TDoWhile.getSize;
var groesse,grt : TSize;
begin
  groesse := getBedSize(cv);
  grt := Block[1].getSize(cv,Options);
  groesse.cy := groesse.cy+grt.cy+cv.TextExtent('ja').cy+14;
  if typDoW then
    groesse.cx := groesse.cx + + 2*cv.TextExtent('ja').cx - 4
  else
    groesse.cx := groesse.cx + + 2*cv.TextExtent('nein').cx - 4;

  if (grt.cx > groesse.cx) then groesse.cx := grt.cx+18;
  result := groesse;
end;

function TDoWhile.GetBlockPos;
begin
  result := Point(pos.x,pos.y+12);
end;

// --------------------------
//         TFor
// --------------------------

constructor TFor.Create(b: String);
begin
  inherited;
  Block[1].Name := 'For-Block';
end;

procedure TFor.Speichern(f:TFileStream);
var s:array[0..100] of char;
begin
  strcopy(s,'F:');
  strcat(s,PChar(name));
  strcat(s,#1);
  f.WriteBuffer(s,strlen(s));
  block[1].Speichern(f);
  strcopy(s,#2);
  strcat(s,#2);
  f.WriteBuffer(s,2);
  if (next<> nil) then next.Speichern(f);
end;

procedure TFor.Exportieren;
var s:array[0..100] of char;
    text1,text2,text3: string;
    i : integer;

begin
  if (name='???') then
  begin
    Text1 := '?';
    Text2 := '?';
    Text3 := '?';
  end
  else
  begin
    Text1 := '';
    Text2 := '';
    Text3 := '';
    i := 1;
    while (i<=Length(name)) and (name[i]<>' ') do
    begin
      Text1 := Text1 + name[i];
      i := i +1;
    end;
    i := i +4;
    while (i<=Length(name)) and (name[i]<>' ') do
    begin
      Text2 := Text2 + name[i];
      i := i +1;
    end;
    i := i +5;
    while (i<=Length(name)) do
    begin
      Text3 := Text3 + name[i];
      i := i +1;
    end;
  end;
  strcopy(s,PChar(einr));
  strcat(s,'for ');
  strcat(s,PChar(text1));
  strcat(s,':=');
  strcat(s,PChar(text2));
  strcat(s,' to ');
  strcat(s,PChar(text3));
  strcat(s,' do');
  strcat(s,#13+#10);
  strcat(ss,s);
  block[1].exportieren(ss,einr);
  strcopy(s,';');
  strcat(s,#13+#10);
  strcat(ss,s);
  if (next<> nil) then next.Exportieren(ss,einr);
end;

end.




